local addon, private = ...
local mod = LibStub("AceAddon-3.0"):NewAddon(addon, "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")
local ArtifactUI = _G.C_ArtifactUI

function getSocketSlotID()
  local class = select(2, UnitClass("player"))
  if (class == "PALADIN" and GetSpecialization() == 2) or (class == "WARRIOR" and GetSpecialization() == 3) then
    return 17
  else
    return 16
  end
end

function mod:Update()
  self:BuildOutput()
  self.CopyFrame:Show()
end

function mod:copySlotsFromEquipped()
  for _, frame in ipairs(self.frameSlots) do
    local item = nil
    if frame.info.artifact_gem_id then
      local _, gemLink = GetItemGem(GetInventoryItemLink('player', getSocketSlotID()), frame.info.artifact_gem_id)
      item = gemLink
    else
      item = GetInventoryItemLink('player', frame.info.slot_id)
    end
    frame.setItem(item, item)
  end
  self:Update()
end

function mod:OnInitialize()
  mod:RegisterChatCommand('sim', self.loadWindow)
end

function mod:BuildTalentString()
  local talentInfo = {}
  local maxTiers = 7
  local maxColumns = 3
  for tier = 1, maxTiers do
    for column = 1, maxColumns do
      local talentID, name, iconTexture, selected, available = GetTalentInfo(tier, column, GetActiveSpecGroup())
      if selected then
        talentInfo[tier] = column
      end
    end
  end

  local str = 'talents='
  for i = 1, maxTiers do
    if talentInfo[i] then
      str = str .. talentInfo[i]
    else
      str = str .. '0'
    end
  end

  return str
end

local lastArtifactString = nil
function mod:BuildArtifactString()
  local item_id = select(1, ArtifactUI.GetArtifactInfo())
  if item_id == nil or item_id == 0 then
    return lastArtifactString
  end

  local artifact_id = self.ARTIFACT_TABLE[item_id]
  if artifact_id == nil then
    return lastArtifactString
  end

  -- Note, relics are handled by the item string
  local str = 'artifact=' .. artifact_id .. ':0:0:0:0'

  local powers = ArtifactUI.GetPowers()
  for i = 1, #powers do
    local powerId = powers[i]
    local powerInfo = ArtifactUI.GetPowerInfo(powerId)
    if powerInfo.currentRank > 0 and powerInfo.currentRank - powerInfo.bonusRanks > 0 then
      str = str .. ':' .. powerId .. ':' .. (powerInfo.currentRank - powerInfo.bonusRanks)
    end
  end

  lastArtifactString = str
  return str
end

local function getItemSplit(itemLink)
  local itemString = string.match(itemLink, "item:([%-?%d:]+)")
  local itemSplit = {}

  -- Split data into a table
  for v in string.gmatch(itemString, "(%d*):?") do
    itemSplit[#itemSplit + 1] = tonumber(v) or 0
  end
  return itemSplit
end

local function getEnchantID(item, split)
  split = split or getItemSplit(item)
  return split[mod.OFFSETS.ENCHANT_ID]
end

local function getFlagValues(itemSplit)
  local flags = itemSplit[mod.OFFSETS.FLAGS]
  local rest_offset = mod.OFFSETS.BONUS_ID + itemSplit[mod.OFFSETS.BONUS_ID] + 1
  local upgrade, drop_level = nil, nil, nil

  -- Upgrade level
  if bit.band(flags, 4) == 4 then
    local upgrade_id = tonumber(itemSplit[rest_offset])
    if mod.UPGRADE_TABLE[upgrade_id] ~= nil and mod.UPGRADE_TABLE[upgrade_id] > 0 then
      upgrade = mod.UPGRADE_TABLE[upgrade_id]
    end
    rest_offset = rest_offset + 1
  end

  -- Artifacts use this
  local equippedRelics, selectedRelics = nil, nil
  if bit.band(flags, 0x100) == 0x100 then

    -- fully-filled out artifacts have an extra field
    if bit.band(flags, 0x1000000) == 0x1000000 then
      rest_offset = rest_offset + 1
    end

    equippedRelics = {}
    selectedRelics = {}
    rest_offset = rest_offset + 1
    local relic_str = ''
    while rest_offset < #itemSplit do
      local n_bonus_ids = itemSplit[rest_offset]
      rest_offset = rest_offset + 1

      local relic_ids = {}
      if n_bonus_ids == 0 then
        tinsert(relic_ids, 0)
      else
        for rbid = 1, n_bonus_ids do
          tinsert(relic_ids, itemSplit[rest_offset])
          rest_offset = rest_offset + 1
        end
      end
      tinsert(equippedRelics, table.concat(relic_ids, ":"))
    end

    for i = 1, #equippedRelics do
      local prospectiveLink = mod.frameLookup["relic_" .. i].itemLink
      if prospectiveLink then
        local bonuses = {}
        local prospectiveSplit = getItemSplit(mod.frameLookup["relic_" .. i].itemLink)
        if prospectiveSplit[mod.OFFSETS.BONUS_ID] == 0 then
          tinsert(bonuses, 0)
        else
          for index=1, prospectiveSplit[mod.OFFSETS.BONUS_ID] do
            bonuses[#bonuses + 1] = prospectiveSplit[mod.OFFSETS.BONUS_ID + index]
          end
        end
        tinsert(selectedRelics, table.concat(bonuses, ":"))
      end
    end
  end

  -- Some leveling quest items seem to use this, it'll include the drop level of the item
  if bit.band(flags, 512) == 512 then
    rest_offset = rest_offset + 1
    drop_level = itemSplit[rest_offset]
  end

  return upgrade, equippedRelics, selectedRelics, drop_level
end

function mod:BuildSimcLine(itemLink, equippedItem, useEquippedRelics)
  local itemName = GetItemInfo(itemLink)
  local itemSplit = getItemSplit(itemLink)
  local simcItemOptions = {}
  local slug_name = gsub(gsub(string.lower(itemName), " ", "_"), "[^%a_]", "")
  tinsert(simcItemOptions, slug_name)

  local itemType = select(9, GetItemInfo(itemLink))
  local isArtifactWeapon = ((itemType == "INVTYPE_WEAPONMAINHAND" or itemType == "INVTYPE_2HWEAPON") and getSocketSlotID() == INVSLOT_MAINHAND) or (itemType == "INVTYPE_SHIELD" and getSocketSlotID() == INVSLOT_OFFHAND)

  -- Item id
  local itemId = itemSplit[self.OFFSETS.ITEM_ID]
  simcItemOptions[#simcItemOptions + 1] = 'id=' .. itemId

    -- Enchant
  if getEnchantID(nil, itemSplit) > 0 then
    simcItemOptions[#simcItemOptions + 1] = 'enchant_id=' .. getEnchantID(nil, itemSplit)
  elseif equippedItem and getEnchantID(equippedItem) > 0 then
    simcItemOptions[#simcItemOptions + 1] = 'enchant_id=' .. getEnchantID(equippedItem)
  end

  -- New style item suffix, old suffix style not supported
  if tonumber(itemSplit[self.OFFSETS.SUFFIX_ID]) ~= 0 then
    simcItemOptions[#simcItemOptions + 1] = 'suffix=' .. itemSplit[self.OFFSETS.SUFFIX_ID]
  end

  local flags = itemSplit[self.OFFSETS.FLAGS]

  local upgrade, equippedRelics, selectedRelics, drop_level = getFlagValues(itemSplit, useEquippedRelics)
  if upgrade then
    simcItemOptions[#simcItemOptions + 1] = 'upgrade=' .. upgrade
  end

  -- This is hacky, but ensures we don't build relic strings for non-primary artifact weapons (ie, rogue offhands, paladin mainhands)
  if not isArtifactWeapon then
    if equippedRelics then
      for i=1, #equippedRelics do
        equippedRelics[i] = 0
      end
    end
    if selectedRelics then
      for i=1, #selectedRelics do
        selectedRelics[i] = 0
      end
    end
  end

  if useEquippedRelics and equippedRelics and #equippedRelics > 0 then
    simcItemOptions[#simcItemOptions + 1] = 'relic_id=' .. table.concat(equippedRelics, "/")
  elseif selectedRelics and #selectedRelics > 0 then
    simcItemOptions[#simcItemOptions + 1] = 'relic_id=' .. table.concat(selectedRelics, "/")
  end

  if drop_level then
    simcItemOptions[#simcItemOptions + 1] = 'drop_level=' .. drop_level
  end

  local bonuses = {}
  for index=1, itemSplit[self.OFFSETS.BONUS_ID] do
    bonuses[#bonuses + 1] = itemSplit[self.OFFSETS.BONUS_ID + index]
  end
  if #bonuses > 0 then
    simcItemOptions[#simcItemOptions + 1] = 'bonus_id=' .. table.concat(bonuses, '/')
  end

  -- Gems
  local gems = {}

  -- Get sockets for artifacts
  if isArtifactWeapon then
    for _, frame in ipairs(mod.frameSlots) do
      if frame.itemLink and frame.info.artifact_gem_id then
        local gemLink = useEquippedRelics and frame.equippedItem or frame.itemLink
        if gemLink then
          local gemDetail = string.match(gemLink, "item[%-?%d:]+")
          gems[#gems + 1] = string.match(gemDetail, "item:(%d+):" )
        elseif flags == 256 then
          gems[#gems + 1] = "0"
        end
      end
    end

  -- Sockets for non-artifacts
  else
    for i=1, 4 do -- hardcoded here to just grab all 4 sockets
      local _, gemLink = GetItemGem(itemLink, i)
      if gemLink then
        local gemDetail = string.match(gemLink, "item[%-?%d:]+")
        gems[#gems + 1] = string.match(gemDetail, "item:(%d+):" )
      elseif flags == 256 then
        gems[#gems + 1] = "0"
      end
    end
  end


  if #gems > 0 then
    simcItemOptions[#simcItemOptions + 1] = 'gem_id=' .. table.concat(gems, '/')
  end

  return table.concat(simcItemOptions, ',')
end

function mod:BuildOutput()
  local text = ""

  if mod.mode == "compare" then
    text = text .. "# Paste into a SimC profile to compare your current loadout against a proposed loadout\n"
    local profile, loadout = self:BuildLoadout("equipped", {})
    local profile_name = UnitName("player") .. ' - ' .. select(2, GetSpecializationInfo(GetSpecialization())).. ' (Equipped)'
    text = text .. '\n' .. strlower(select(2, UnitClass('player'))) .. '="' .. profile_name .. '"\n'
    local profile2, _ = self:BuildLoadout("selected", loadout)
    text = text .. profile
    if #profile2 > 0 then
      text = text .. '\ncopy="' .. UnitName("player") .. ' - ' .. select(2, GetSpecializationInfo(GetSpecialization())).. ' (Comparison)","' .. profile_name .. '"\n'
      text = text .. profile2
    end
  elseif mod.mode == "build" then
    text = text .. "# Paste into your SimC profile to sim this loadout\n"
    text = text .. self:BuildLoadout("selected", {})
  end
  mod.editBox:SetText(text .. "\n")
  mod.CopyFrame:Show()
end

function mod:BuildLoadout(source, existingItems)
  local text = ""
  local loadout = {}
  local talents = self:BuildTalentString() .. "\n"
  local artifact = (self:BuildArtifactString() or lastArtifactString or "").. "\n"
  local race = strlower(UnitRace("player")):gsub(" ", "_")
  if race == "night_elf" then
    race = race .. "\ntimeofday=day"
  end
  local level = tostring(UnitLevel("player"))
  local spec, role = unpack(mod.SpecInfo[GetSpecializationInfo(GetSpecialization())])
  local position = role == "tank" and "front" or "back"

  spec = spec:gsub(" ", "_")  -- "beast mastery -> beast_mastery"
  if existingItems.talents ~= talents then text = text .. talents end
  if existingItems.artifact ~= artifact then text = text .. artifact end
  if existingItems.race ~= race then text = text .. "race=" .. race .. "\n" end
  if existingItems.level ~= level then text = text .. "level=" .. level .. "\n" end
  if existingItems.role ~= role then text = text .. "role=" .. role .. "\n" end
  if existingItems.spec ~= spec then text = text .. "spec=" .. spec .. "\n" end
  if existingItems.position ~= position then text = text .. "position=" .. position .. "\n" end

  loadout.artifact = artifact
  loadout.talents = talents
  -- todo
  loadout.race = race
  loadout.level = level
  loadout.spec = spec
  loadout.role = role
  loadout.position = position

  local finger, trinket, relicNames = 1, 1, {}
  for _, frame in ipairs(mod.frameSlots) do
    local name, itemLink, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice, classId, subClassId, link
    if source == "selected" and frame.itemLink then
      link = frame.itemLink
    elseif source == "equipped" then
      link = frame.equippedItem
    end

    if link then
      name, itemLink, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice, classId, subClassId = GetItemInfo(link)
    end
    local key = tostring(frame.info.slot_id) .. (name or "no item")

    if frame.info.slot_id and not frame.info.artifact_gem_id and link then
      local line = self:BuildSimcLine(link, frame.equippedItem, source == "equipped")
      if existingItems[key] ~= line then
        text = text .. frame.info.slot_name .. "=" .. line .. "\n"
        loadout[key] = line
      end
    elseif frame.info.artifact_gem_id then
      local existing
      if frame.equippedItem then
        existing = GetItemInfo(frame.equippedItem)
      end
      if name or existing then
        tinsert(relicNames, name or existing)
      end
    end
  end
  if #relicNames > 0 then
    local relics = "# Relics: " .. table.concat(relicNames, ', ') .. "\n"
    if existingItems.relics ~= relics then
      text =  text .. relics
    end
    loadout.relics = relics
  end
  return text, loadout
end
